from .hash import HashTrait
from .append import AppendTrait
from .slice import SliceTrait
from .canonicalize import CanonicalizeTrait

__all__ = [
    "HashTrait",
    "AppendTrait",
    "SliceTrait",
    "CanonicalizeTrait",
]
